import numpy as np
import pytest
from test_manager import temp_manager


@pytest.mark.parametrize(
    "data",
    [
        np.array([1, 2, 3]),
        np.array([1.2, 2.3, 3.4]),
        np.array([[1, 2], [3, 4]]),
    ],
)
def test_read_field(temp_manager, data):
    sim = temp_manager.create_simulation()
    uid = sim.uid
    sim.add_field("test", data)
    sim.finish_step()

    sim = temp_manager.sim(uid)
    data_read = sim.data["test"].at_step(0)
    assert np.array_equal(data_read, data)



# This file is part of bamboost, a Python library built for datamanagement
# using the HDF5 file format.
#
# https://gitlab.ethz.ch/compmechmat/research/libs/dbmanager
#
# Copyright 2023 Flavio Lorez and contributors
#
# There is no warranty for this code

"""Utility functions used by bamboost.
"""

from collections.abc import MutableMapping
from itertools import islice
from pathlib import Path

import h5py
import pandas as pd

__all__ = ["flatten_dict", "unflatten_dict", "tree", "h5_tree"]

space = "    "
branch = "│   "
tee = "├── "
last = "└── "


def flatten_dict(dictionary, parent_key="", seperator="."):
    items = []
    for key, value in dictionary.items():
        new_key = parent_key + seperator + key if parent_key else key
        if isinstance(value, MutableMapping):
            items.extend(flatten_dict(value, new_key, seperator=seperator).items())
        else:
            items.append((new_key, value))
    return dict(items)


def unflatten_dict(dictionary, seperator="."):
    new_dict = dict()
    for key, value in dictionary.items():
        parts = key.split(seperator)
        d = new_dict
        for part in parts[:-1]:
            if part not in d:
                d[part] = dict()
            d = d[part]
        d[parts[-1]] = value
    return new_dict


# https://stackoverflow.com/questions/9727673/list-directory-tree-structure-in-python
def tree(
    dir_path: Path,
    level: int = -1,
    limit_to_directories: bool = False,
    length_limit: int = 1000,
):
    """Given a directory Path object print a visual tree structure"""
    dir_path = Path(dir_path)  # accept string coerceable to Path
    files = 0
    directories = 0
    folder_symbol = "\U00002B57 "

    def inner(dir_path: Path, prefix: str = "", level=-1):
        nonlocal files, directories
        if not level:
            return  # 0, stop iterating
        if limit_to_directories:
            contents = [d for d in dir_path.iterdir() if d.is_dir()]
        else:
            contents = list(dir_path.iterdir())
        pointers = [tee] * (len(contents) - 1) + [last]
        for pointer, path in zip(pointers, contents):
            if path.is_dir():
                yield prefix + pointer + "\U000025CC " + path.name
                directories += 1
                extension = branch if pointer == tee else space
                yield from inner(path, prefix=prefix + extension, level=level - 1)
            elif not limit_to_directories:
                yield prefix + pointer + path.name
                files += 1

    tree_string = ""
    tree_string += (folder_symbol + dir_path.name) + "\n"
    iterator = inner(dir_path, level=level)
    for line in islice(iterator, length_limit):
        tree_string += (line) + "\n"
    if next(iterator, None):
        tree_string += (f"... length_limit, {length_limit}, reached, counted:") + "\n"
    tree_string += (
        f"\n{directories} directories" + (f", {files} files" if files else "")
    ) + "\n"
    return tree_string


def h5_tree(val, pre=""):
    items = len(val)
    for key, val in val.items():
        items -= 1
        if items == 0:
            # the last item
            if type(val) == h5py._hl.group.Group:
                print(pre + "└── " + key)
                h5_tree(val, pre + "    ")
            else:
                print(pre + "└── " + key + " (%d)" % len(val))
        else:
            if type(val) == h5py._hl.group.Group:
                print(pre + "├── " + key)
                h5_tree(val, pre + "│   ")
            else:
                print(pre + "├── " + key + " (%d)" % len(val))


def show_differences(df: pd.DataFrame) -> pd.DataFrame:
    """This function takes a pandas DataFrame as input and returns a modified
    DataFrame that shows only the columns which have differences.

    Args:
        - df (pd.DataFrame): The input DataFrame to analyze

    Returns:
        - pd.DataFrame

    The function first creates a copy of the input DataFrame to work with. It
    then iterates over each column in the DataFrame and tries to calculate the
    number of unique values in that column. If successful, it adds the column
    name and number of unique values to a list of good results. If there is an
    error, it attempts to apply json.dumps to the column and then calculate the
    number of unique values again. If this is successful, it also adds the
    column name and number of unique values to the list of good results. If
    there is still an error, it adds the column name and the error to a list of
    errors.

    After processing all columns, the function removes any columns that had
    errors from the DataFrame. It then sets the index of the DataFrame to 'id'
    and filters out any columns that have only one unique value. The modified
    DataFrame is then returned.
    """
    import json

    df_diff = df.copy()
    cols_nunique_good = []
    cols_nunique_error = []
    for col in df_diff.columns:
        try:
            nunique = df_diff[col].nunique()
            cols_nunique_good.append((col, nunique))
        except Exception as e:
            try:
                df_diff[col] = df_diff[col].apply(json.dumps)
                nunique = df_diff[col].nunique()
                cols_nunique_good.append((col, nunique))
            except TypeError as e:
                cols_nunique_error.append((col, e))

    df_diff = df_diff[
        df_diff.columns[~df_diff.columns.isin([col for col, _, in cols_nunique_error])]
    ]
    try:
        df_diff.set_index("id", inplace=True)
    except KeyError:
        pass
    df_diff = df_diff.loc[:, (df_diff.nunique() != 1)]
    df_diff.dropna(axis=1, how="all", inplace=True)
    return df_diff


def to_camel_case(s: str) -> str:
    words = s.split()
    camel_case = words[0].lower() + "".join([word.capitalize() for word in words[1:]])
    return camel_case



# This file is part of bamboost, a Python library built for datamanagement
# using the HDF5 file format.
#
# https://gitlab.ethz.ch/compmechmat/research/libs/dbmanager
#
# Copyright 2023 Flavio Lorez and contributors
#
# There is no warranty for this code

from __future__ import annotations

import logging
import os
import pkgutil
import sqlite3
import subprocess

import pandas as pd

from bamboost._config import config
from bamboost._sqlite_database import SQLiteHandler, with_connection
from bamboost.common.mpi import MPI
from bamboost.index import DatabaseTable, IndexAPI
from bamboost.manager import Manager

log = logging.getLogger(__name__)

HOME = os.path.expanduser("~")
CACHE_DIR = os.path.join(HOME, ".cache", "bamboost")
# location off the index file on the remote server
REMOTE_INDEX = f"~/.local/share/bamboost/bamboost.db"


class Remote(IndexAPI, SQLiteHandler):
    """Access bamboost database of a remote server. The index is fetched using
    rsync over ssh. The `remote_name` can be a hostname or an IP address. Make
    sure that ssh keys are set and working, as there is no user authentication.
    The `skip_update` flag can be set to avoid fetching the index from the
    remote server.

    Args:
        - remote_name (str): The hostname or IP address of the remote server.
        - skip_update (bool): Flag to avoid fetching the index from the remote
          server. Default is False.

    Example:
        >>> remote = Remote("euler")
        >>> remote.read_table()
        returns a pandas DataFrame of the remote index.
        >>> remote["<id>"]
        returns a RemoteManager object for the given id.
    """

    def __new__(cls, *args, **kwargs):
        """Override the __new__ method to avoid the singleton pattern of IndexAPI."""
        return object.__new__(cls)

    def __init__(self, remote_name: str, skip_update: bool = False) -> None:
        self.remote_name = remote_name
        self.local_path = os.path.join(CACHE_DIR, self.remote_name)
        os.makedirs(self.local_path, exist_ok=True)
        self.file = f"{self.local_path}/bamboost.db"
        if not skip_update:
            self._fetch_index()

        # Initialize the SQLiteHandler
        SQLiteHandler.__init__(self, self.file)

    def _fetch_index(self) -> None:
        """Fetches the bamboost.db from the remote server. Takes time."""
        subprocess.run(
            ["rsync", "-avh", f"{self.remote_name}:{REMOTE_INDEX}", f"{self.file}"]
        )

    def _ipython_key_completions_(self) -> list[str]:
        ids = self.read_table()[["id", "path"]].values
        completion_keys = [
            f'{key} - {"..."+val[-25:] if len(val)>=25 else val}' for key, val in ids
        ]
        return completion_keys

    def __getitem__(self, id: str) -> RemoteManager:
        """Return a `RemoteManager` for the given id."""
        id = id.split(" - ")[0]
        return RemoteManager(id, remote=self)

    def get_manager(self, id: str, skip_update: bool = False) -> RemoteManager:
        return RemoteManager(id, remote=self, skip_update=skip_update)

    @with_connection
    def get_path(self, id: str) -> str:
        self._cursor.execute("SELECT path FROM dbindex WHERE id=?", (id,))
        return self._cursor.fetchone()[0]

    @with_connection
    def insert_local_path(self, id: str, path: str) -> None:
        try:
            self._cursor.execute(
                "UPDATE dbindex SET local_path=? WHERE id=?", (path, id)
            )
        except sqlite3.OperationalError:
            self._cursor.execute(
                "ALTER TABLE dbindex ADD COLUMN local_path TEXT DEFAULT NULL"
            )
            self._cursor.execute(
                "UPDATE dbindex SET local_path=? WHERE id=?", (path, id)
            )


class RemoteDatabaseTable(DatabaseTable):

    def sync(self) -> None:
        """Don't sync a remote database."""
        return None


class RemoteManager(Manager):
    """
    Manager class with remote functionality. Constructor takes an existing ID
    of a database on a remote server. The ssh connection must be set up to work
    without explicit user authentication. Data is lazily transferred using
    rsync.
    """

    # INFO: The initialization of this class takes around 8 seconds because the
    # home directory of the remote server needs to be fetched, read database
    # index, and get remote database csv file with db.df in it (see below).
    def __init__(
        self,
        id: str,
        remote: Remote,
        comm: MPI.Comm = MPI.COMM_WORLD,
    ) -> None:
        """
        params
        skip_update: if True, does not lookup the new database on the remote:
        """
        self.UID = id
        self.remote = remote
        self.comm = comm
        self.path = os.path.join(self.remote.local_path, self.UID)
        log.info(f"Creating cache directory at {self.path}")
        os.makedirs(self.path, exist_ok=True)

        self.remote_path_db = self._index.get_path(self.UID)

        # check if path exists
        if not os.path.isdir(self.path):
            os.makedirs(self.path)

        self.UID = id
        self._index.insert_local_path(self.UID, self.path)

        # Update the SQL table for the database
        with self._index.open():
            self._table.create_database_table()
            # self._table.sync()

    def _repr_html_(self) -> str:
        html_string = pkgutil.get_data(__name__, "../html/manager.html").decode()
        icon = pkgutil.get_data(__name__, "../html/icon.txt").decode()
        return (
            html_string.replace("$ICON", icon)
            .replace("$db_path", self.path)
            .replace("$db_uid", "euler-" + self.UID)
            .replace("$db_size", str(len(self)))
        )

    def _get_uids(self) -> list:
        """Override the simulation list to fetch from databasetable."""
        return self._index.fetch(f"SELECT id FROM db_{self.UID}")

    @property
    def _index(self) -> IndexAPI:
        return self.remote

    @property
    def _table(self) -> RemoteDatabaseTable:
        return RemoteDatabaseTable(self.UID, _index=self._index)

    def get_view(self, include_linked_sims: bool = False) -> pd.DataFrame:
        df = super().get_view(include_linked_sims)
        df.insert(1, "cached", False)
        for id in os.listdir(self.path):
            if id in df["id"].values:
                df.loc[df["id"] == id, "cached"] = True

        opts = config.get("options", {})
        if "sort_table_key" in opts:
            df.sort_values(
                opts.get("sort_table_key", "id"),
                ascending=opts.get("sort_table_order", "asc") == "asc",
                inplace=True,
            )
        return df

    def sim(self, uid, return_writer: bool = False):
        """Return simulation object.

        Args:
            - uid (str): the unique id of the sim to be transferred.
            - return_writer (bool): Flag to indicate whether to return a writer
              object. Default is False.

        This method checks if the data for the given uid is already in the
        local cache. If not, it transfers the data from a remote location using
        rsync. The method then calls the superclass method to perform further
        operations on the transferred data.
        """
        # Check if data is already in cache
        if os.path.exists(f"{self.path}/{uid}"):
            log.info(f"Data for {uid} already in cache")
            return super().sim(uid, return_writer)

        # Transfer data using rsync
        log.info(f"Data not in cache. Transferring data for {uid} from {self.remote}")
        subprocess.call(
            [
                "rsync",
                "-r",
                f"{self.remote.remote_name}:{self.remote_path_db}/{uid}",
                f"{self.path}",
            ],
            stdout=subprocess.PIPE,
        )
        log.info(f"Data for {uid} transferred to {self.path}")

        return super().sim(uid, return_writer)

    def sync(self, uid: str) -> None:
        subprocess.call(
            [
                "rsync",
                "-r",
                f"{self.remote.remote_name}:{self.remote_path_db}/{uid}",
                f"{self.path}",
            ],
            stdout=subprocess.PIPE,
        )
        log.info(f"Data for {uid} transferred to {self.path}")


if __name__ == "__main__":
    pass



# test_manager.py

import os
import shutil
import tempfile

import numpy as np
import pytest

from bamboost import Manager, Simulation, index


@pytest.fixture()
def temp_manager():
    temp_dir = tempfile.mkdtemp()
    db = Manager(path=temp_dir)
    yield db
    try:
        shutil.rmtree(temp_dir)
    except FileNotFoundError:
        pass


def test_if_new_manager_created():
    temp_dir = tempfile.mkdtemp()
    db = Manager(temp_dir)
    assert os.path.isdir(temp_dir)


class TestManager:

    def test_create_simulation_creates_new_file(self, temp_manager: Manager):
        sim_uid = "this"
        writer = temp_manager.create_simulation(uid=sim_uid)
        h5_file = os.path.join(temp_manager.path, sim_uid, f"{sim_uid}.h5")
        assert os.path.isfile(h5_file)

    @pytest.mark.parametrize("fix_df", [True, False])
    def test_manager_length(self, temp_manager: Manager, fix_df: bool):
        temp_manager.FIX_DF = fix_df
        temp_manager.create_simulation()
        temp_manager.create_simulation()
        temp_manager.create_simulation()
        assert len(temp_manager) == 3

    def test_get_single_simulation_base_case(self, temp_manager: Manager):
        sim_uid = "this"
        temp_manager.create_simulation()
        temp_manager.create_simulation(sim_uid)
        db = Manager(temp_manager.path)
        sim = db.sim(sim_uid)
        assert isinstance(sim, Simulation)
        assert sim.path == os.path.join(temp_manager.path, sim_uid)


# --------------------------------------------------------------------------
# Database syncing between SQLite and Filesystem
# --------------------------------------------------------------------------
def test_if_table_created(temp_manager: Manager):
    uid = temp_manager.UID
    table_name = f"db_{uid}"
    res = index.IndexAPI().fetch("SELECT name FROM sqlite_master WHERE type='table';")
    # assert that the database table exists
    assert (table_name,) in res
    # assert that the table with update times exists
    assert (f"db_{uid}_t",) in res


def test_dataframe_integrity(temp_manager: Manager):
    _index = index.IndexAPI()  # shortcutting calling IndexAPI() every time
    booleans = {
        "1": True,
        "2": False,
        "3": True,
    }

    @_index.commit_once
    def create_sims(booleans: list):
        for args in zip([1, 2, 3], booleans, ["a", "b", "c"]):
            sim = temp_manager.create_simulation(
                uid=args[1],
                parameters={
                    "int": args[0],
                    "float": 1.0,
                    "str": args[2],
                    "boolean": booleans[args[1]],
                    "boolean2": False,
                    "array": np.array([1, 2, 3]),
                },
            )

    create_sims(booleans)

    # delete the table to force re-creation -> booleans are wrong!!!
    with _index.open():
        _index._cursor.execute(f"DROP TABLE db_{temp_manager.UID};")
        _index._cursor.execute(f"DROP TABLE db_{temp_manager.UID}_t;")
        _index.commit()

    db = Manager(uid=temp_manager.UID)
    df = db.df.set_index("id")

    for uid, val in booleans.items():
        assert df.loc[uid, "boolean"] == val



# This file is part of bamboost, a Python library built for datamanagement
# using the HDF5 file format.
#
# https://gitlab.ethz.ch/compmechmat/research/libs/dbmanager
#
# Copyright 2023 Flavio Lorez and contributors
#
# There is no warranty for this code

import argparse
import os

from bamboost._config import paths

__all__ = ["Job"]


class Job:
    def __init__(self):
        pass

    def create_sbatch_script(
        self,
        commands: list,
        path: str,
        uid: str = None,
        db_id: str = None,
        nnodes: int = 1,
        ntasks: int = 4,
        ncpus: int = 1,
        time: str = "04:00:00",
        mem_per_cpu: int = 2048,
        tmp: int = 8000,
    ) -> None:
        """Write sbatch script for new simulation."""
        nb_tasks_per_node = int(ntasks / nnodes)

        # define how mpirun is called
        mpicommand = ""
        if ntasks > 1:
            mpicommand = "mpirun "

        script = f"#!/bin/bash\n\n"

        # sbatch commands
        script += f"#SBATCH --ntasks={ntasks}\n"
        if ntasks > 1:
            script += f"#SBATCH --nodes={nnodes}\n"
            script += f"#SBATCH --cpus-per-task={ncpus}\n"
            script += f"#SBATCH --ntasks-per-node={nb_tasks_per_node}\n"
        script += f"#SBATCH --time={time}\n"
        script += f"#SBATCH --job-name={uid}\n"
        script += f"#SBATCH --mem-per-cpu={mem_per_cpu}\n"
        if tmp:
            script += f"#SBATCH --tmp={tmp}\n"
        script += f"#SBATCH --output={os.path.join(path, uid)}/{uid}.out\n"

        # add SCRIPT_DIR as environment variable
        script += f"""path=$(sqlite3 {paths['DATABASE_FILE']} "SELECT path FROM dbindex WHERE id='{db_id}'")\n"""
        script += f"SCRIPT_DIR=$path/{uid}\n\n"

        # user defined commands
        script += "\n"
        for cmd in commands:
            script += cmd.format(MPI=mpicommand) + "\n"

        # write to submission file
        with open(
            os.path.join(os.path.join(path, uid), f"sbatch_{uid}.sh"), "w"
        ) as file:
            file.write(script)

    def create_bash_script_local(
        self, commands: list, path: str, uid: str, ntasks: int = 4
    ):
        """Write bash script for local execution."""

        # define how mpirun is called
        mpicommand = ""
        if ntasks > 1:
            mpicommand = f"mpirun -n {ntasks}"

        script = f"#!/bin/bash\n\n"

        # add SCRIPT_DIR as environment variable
        script += 'SCRIPT_DIR=$( cd -- "$( dirname -- "${BASH_SOURCE[0]}" )" &> /dev/null && pwd )\n\n'

        # user defined commands
        for cmd in commands:
            script += cmd.format(MPI=mpicommand) + "\n"

        with open(os.path.join(os.path.join(path, uid), f"{uid}.sh"), "w") as file:
            file.write(script)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("file_to_run", type=str)
    # parser.add_argument('input_file', type=str)
    parser.add_argument("--uid", type=str, default=None)
    parser.add_argument("--nodes", type=int, default=1)
    parser.add_argument("--ntasks", type=int, default=4)
    parser.add_argument("--ncpus", type=int, default=1)
    parser.add_argument("--time", type=str, default="04:00:00")
    parser.add_argument("--mem_per_cpu", type=int, default=2048)
    args = parser.parse_args()

    job = Job()
    job.create_sbatch_script(
        [args.file_to_run],
        args.uid,
        args.nodes,
        args.ntasks,
        args.ncpus,
        args.time,
        args.mem_per_cpu,
    )



__author__ = "florez@ethz.ch"
__copyright__ = ""
__license__ = "LGPLv3"
__version__ = "0.6.0"

import logging
import os

from .manager import Manager
from .simulation import Simulation
from .simulation_writer import SimulationWriter
from ._config import config
from .extensions import extensions


def set_log_level(level: int = 30):
    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)
    logging.basicConfig(level=os.environ.get("LOGLEVEL", level))


# Set default logging level to WARNING (30)
set_log_level(30)



# This file is part of bamboost, a Python library built for datamanagement
# using the HDF5 file format.
#
# https://gitlab.ethz.ch/compmechmat/research/libs/dbmanager
#
# Copyright 2023 Flavio Lorez and contributors
#
# There is no warranty for this code

from __future__ import annotations

import logging
import os
import pkgutil
import subprocess
from typing import Any, Iterable, Tuple

import numpy as np
import pandas as pd
from typing_extensions import Self, deprecated

from bamboost._config import config

from . import index
from .accessors.fielddata import DataGroup
from .accessors.globals import GlobalGroup
from .accessors.meshes import MeshGroup
from .common import hdf_pointer, utilities
from .common.file_handler import FileHandler, with_file_open
from .common.job import Job
from .common.mpi import MPI
from .xdmf import XDMFWriter

__all__ = ["Simulation", "Links"]

log = logging.getLogger(__name__)


class Links(hdf_pointer.MutableGroup):
    """Link group. Used to create and access links.

    I don't know how to distribute this to its own file in the accessors
    directory, due to circular imports.
    """

    def __init__(self, file_handler: FileHandler) -> None:
        super().__init__(file_handler, path_to_data="links")

    def _ipython_key_completions_(self):
        return tuple(self.all_links().keys())

    @with_file_open("r", driver="mpio")
    def __getitem__(self, key) -> Simulation:
        """Returns the linked simulation object."""
        return Simulation.fromUID(self.obj.attrs[key])

    def __setitem__(self, key, newvalue):
        """Creates the link."""
        return self.update_attrs({key: newvalue})

    def __delitem__(self, key):
        """Delete a link."""
        with self._file("a"):
            del self.obj.attrs[key]

    @with_file_open("r")
    def __repr__(self) -> str:
        return repr(
            pd.DataFrame.from_dict(self.all_links(), orient="index", columns=["UID"])
        )

    @with_file_open("r")
    def _repr_html_(self) -> str:
        return pd.DataFrame.from_dict(
            self.all_links(), orient="index", columns=["UID"]
        )._repr_html_()

    @with_file_open("r")
    def all_links(self) -> dict:
        return dict(self.obj.attrs)


class Simulation:
    """A single dataset/simulation. Used to write to it, read from it or append.

    Args:
        uid (str): unique identifier
        path (str): path to parent/database folder
        comm (MPI.Comm): MPI communicator (default=MPI.COMM_WORLD)
    """

    _mesh_location = "Mesh/0"
    _default_mesh = "mesh"

    def __init__(
        self,
        uid: str,
        path: str,
        comm: MPI.Comm = MPI.COMM_WORLD,
        create_if_not_exists: bool = False,
        *,
        _db_id: str = None,
    ):
        self.uid: str = uid
        path = comm.bcast(path, root=0)
        self.path_database: str = os.path.abspath(path)
        self.database_id = _db_id or index.get_uid_from_path(self.path_database)
        self.path: str = os.path.abspath(os.path.join(path, uid))
        self.h5file: str = os.path.join(self.path, f"{self.uid}.h5")
        self.xdmffile: str = os.path.join(self.path, f"{self.uid}.xdmf")

        if not os.path.exists(self.h5file) and not create_if_not_exists:
            raise FileNotFoundError(
                f"Simulation {self.uid} does not exist in {self.path}."
            )

        os.makedirs(self.path, exist_ok=True)

        # MPI information
        self._comm = comm
        self._psize = self._comm.size
        self._prank = self._comm.rank
        self._ranks = np.array([i for i in range(self._psize)])

        self._file = FileHandler(self.h5file)

        # Initialize groups to meshes, data and userdata. Create groups.
        self.meshes: MeshGroup = MeshGroup(self._file)
        self.data: DataGroup = DataGroup(self._file, self.meshes)
        self.globals: GlobalGroup = GlobalGroup(self._file, "/globals")
        self.userdata: hdf_pointer.MutableGroup = hdf_pointer.MutableGroup(
            self._file, "/userdata"
        )
        self.links: Links = Links(self._file)

    @classmethod
    def fromUID(cls, full_uid: str) -> Self:
        """Return the `Simulation` with given UID.

        Args:
            full_uid: the full id (Database uid : simulation uid)
        """
        db_uid, sim_uid = full_uid.split(":")
        db_path = index.IndexAPI().get_path(db_uid)
        return cls(sim_uid, db_path, create_if_not_exists=False)

    @with_file_open()
    def __getitem__(self, key) -> hdf_pointer.BasePointer:
        """Direct access to HDF5 file.

        Returns:
            :class:`~bamboost.common.file_handler.BasePointer`
        """
        return hdf_pointer.BasePointer.new_pointer(self._file, key)

    def _repr_html_(self) -> str:
        html_string = pkgutil.get_data(__name__, "html/simulation.html").decode()
        icon = pkgutil.get_data(__name__, "html/icon.txt").decode()

        table_string = ""
        for key, value in self.parameters.items():
            if (
                isinstance(value, Iterable)
                and not isinstance(value, str)
                and len(value) > 5
            ):
                value = "..."
            table_string += f"""
            <tr>
                <td>{key}</td>
                <td>{value}</td>
            </tr>
            """

        metadata = self.metadata

        def get_pill_div(text: str, color: str):
            return (
                f'<div class="status" style="background-color:'
                f'var(--bb-{color});">{text}</div>'
            )

        status_options = {
            "Finished": get_pill_div("Finished", "green"),
            "Failed": get_pill_div("Failed", "red"),
            "Initiated": get_pill_div("Initiated", "grey"),
        }
        submitted_options = {
            True: get_pill_div("Submitted", "green"),
            False: get_pill_div("Not submitted", "grey"),
        }

        html_string = (
            html_string.replace("$UID", self.uid)
            .replace("$ICON", icon)
            .replace("$TREE", self.show_files(printit=False).replace("\n", "<br>"))
            .replace("$TABLE", table_string)
            .replace("$NOTE", metadata["notes"])
            .replace(
                "$STATUS",
                status_options.get(
                    metadata["status"],
                    f'<div class="status">{metadata["status"]}</div>',
                ),
            )
            .replace("$SUBMITTED", submitted_options[metadata.get("submitted", False)])
            .replace("$TIMESTAMP", metadata["time_stamp"])
        )
        return html_string

    def _push_update_to_sqlite(self, update_dict: dict) -> None:
        """Push update to sqlite database.

        Args:
            - update_dict (dict): key value pair to push
        """
        if not config["options"].get("sync_table", True):
            return
        try:
            index.DatabaseTable(self.database_id).update_entry(self.uid, update_dict)
        except index.Error as e:
            log.warning(f"Could not update sqlite database: {e}")

    @property
    def parameters(self) -> dict:
        tmp_dict = dict()
        if self._prank == 0:
            with self._file("r"):
                # return if parameters is not in the file
                if "parameters" not in self._file.keys():
                    return {}
                tmp_dict.update(self._file["parameters"].attrs)
                for key in self._file["parameters"].keys():
                    tmp_dict.update({key: self._file[f"parameters/{key}"][()]})

        tmp_dict = utilities.unflatten_dict(tmp_dict)

        tmp_dict = self._comm.bcast(tmp_dict, root=0)
        return tmp_dict

    @property
    def metadata(self) -> dict:
        tmp_dict = dict()
        if self._prank == 0:
            with self._file("r") as file:
                tmp_dict.update(file.attrs)

        tmp_dict = utilities.unflatten_dict(tmp_dict)
        tmp_dict = self._comm.bcast(tmp_dict, root=0)
        return tmp_dict

    def files(self, filename: str) -> str:
        """Get the path to the file.

        Args:
            filename: name of the file
        """
        return os.path.join(self.path, filename)

    def show_files(
        self, level=-1, limit_to_directories=False, length_limit=1000, printit=True
    ) -> str:
        """Show the file tree of the simulation directory.

        Args:
            level: how deep to print the tree
            limit_to_directories: only print directories
            length_limit: cutoff
        """
        tree_string = utilities.tree(
            self.path, level, limit_to_directories, length_limit
        )
        if printit:
            print(tree_string)
        else:
            return tree_string

    def open_in_file_explorer(self) -> None:
        """Open the simulation directory. Uses `xdg-open` on linux systems."""
        if os.name == "nt":  # should work on Windows
            os.startfile(self.path)
        else:
            subprocess.run(["xdg-open", self.path])

    def open_in_paraview(self) -> None:
        """Open the xdmf file in paraview."""
        subprocess.Popen(["paraview", self.xdmffile])

    def get_full_uid(self) -> str:
        """Returns the full uid of the simulation (including the one of the database)"""
        database_uid = index.get_uid_from_path(self.path_database)
        return f"{database_uid}:{self.uid}"

    def change_status(self, status: str) -> None:
        """Change status of simulation.

        Args:
            status (str): new status
        """
        if self._prank == 0:
            self._file.open("a")
            self._file.attrs["status"] = status
            self._file.close()

        self._push_update_to_sqlite({"status": status})

    def update_metadata(self, update_dict: dict) -> None:
        """Update the metadata attributes.

        Args:
            update_dict: dictionary to push
        """
        if self._prank == 0:
            update_dict = utilities.flatten_dict(update_dict)
            with self._file("a") as file:
                file.attrs.update(update_dict)

            self._push_update_to_sqlite(update_dict)

    def update_parameters(self, update_dict: dict) -> None:
        """Update the parameters dictionary.

        Args:
            update_dict: dictionary to push
        """
        if self._prank == 0:
            update_dict = utilities.flatten_dict(update_dict)
            with self._file("a") as file:
                file["parameters"].attrs.update(update_dict)

            self._push_update_to_sqlite(update_dict)

    def create_xdmf_file(self, fields: list = None, nb_steps: int = None) -> None:
        """Create the xdmf file to read in paraview.

        Args:
            fields (list[str]): fields for which to write timeseries information,
                if not specified, all fields in data are written.
            nb_steps (int): number of steps the simulation has
        """

        if self._prank == 0:
            with self._file("r") as f:
                if "data" not in f.keys():
                    fields, nb_steps = [], 0
                if fields is None:
                    fields = list(f["data"].keys())

                if nb_steps is None:
                    grp_name = list(f["data"].keys())[0]
                    nb_steps = list(f[f"data/{grp_name}"].keys())
                    nb_steps = max(
                        [int(step) for step in nb_steps if not step.startswith("__")]
                    )

                # temporary fix to load coordinates/geometry
                coords_name = (
                    "geometry"
                    if "geometry"
                    in f[f"{self._mesh_location}/{self._default_mesh}"].keys()
                    else "coordinates"
                )

            xdmf_writer = XDMFWriter(self.xdmffile, self.h5file)
            xdmf_writer.write_points_cells(
                f"{self._mesh_location}/{self._default_mesh}/{coords_name}",
                f"{self._mesh_location}/{self._default_mesh}/topology",
            )

            if fields:
                xdmf_writer.add_timeseries(nb_steps + 1, fields)
            xdmf_writer.write_file()

    def create_batch_script(
        self,
        commands: list = None,
        nnodes=1,
        ntasks=4,
        ncpus=1,
        time="04:00:00",
        mem_per_cpu=2048,
        tmp=None,
        euler=True,
    ) -> None:
        """Create a batch job and put it into the folder.

        Args:
            commands: A list of strings being the user defined commands to run
            nnodes: nb of nodes (default=1)
            ntasks: nb of tasks (default=4)
            ncpus: nb of cpus per task (default=1)
            time: requested time (default=4 hours)
            mem_per_cpu: memory (default=2048)
            tmp: temporary storage, set None to exclude option (default=8000)
            euler: If false, a local bash script will be written
        """
        job = Job()
        if not commands:
            if hasattr(self, "executable"):
                if ".py" in self.executable:
                    command = (
                        f"{{MPI}} python3 {os.path.join(self.path, self.executable)} "
                        f"--path {self.path_database} --uid {self.uid}"
                    )
                    commands = [command]
            else:
                raise AttributeError(
                    """Either you must specify an executable or have it 
                                     copied before with `copy_executable`!"""
                )

        if euler:
            job.create_sbatch_script(
                commands,
                path=os.path.abspath(self.path_database),
                uid=self.uid,
                db_id=self.database_id,
                nnodes=nnodes,
                ntasks=ntasks,
                ncpus=ncpus,
                time=time,
                mem_per_cpu=mem_per_cpu,
                tmp=tmp,
            )
        else:
            job.create_bash_script_local(
                commands,
                path=os.path.abspath(self.path_database),
                uid=self.uid,
                ntasks=ntasks,
            )
        with self._file("a") as file:
            file.attrs.update({"submitted": False})

    def submit(self) -> None:
        """Submit the job for this simulation."""
        if f"sbatch_{self.uid}.sh" in os.listdir(self.path):
            batch_script = os.path.abspath(
                os.path.join(self.path, f"sbatch_{self.uid}.sh")
            )
            env = os.environ.copy()
            _ = env.pop("BAMBOOST_MPI", None)
            subprocess.run(["sbatch", f"{batch_script}"], env=env)
        elif f"{self.uid}.sh" in os.listdir(self.path):
            bash_script = os.path.abspath(os.path.join(self.path, f"{self.uid}.sh"))
            env = os.environ.copy()
            _ = env.pop("BAMBOOST_MPI", None)
            subprocess.run(["bash", f"{bash_script}"], env=env)
        else:
            raise FileNotFoundError(
                f"Could not find a batch script for simulation {self.uid}."
            )
        

        log.info(f"Simulation {self.uid} submitted!")

        with self._file("a") as file:
            file.attrs.update({"submitted": True})

        self._push_update_to_sqlite({"submitted": True})

    @with_file_open("a")
    def change_note(self, note) -> None:
        if self._prank == 0:
            self._file.attrs["notes"] = note
            self._push_update_to_sqlite({"notes": note})

    # Ex-Simulation reader methods
    # ----------------------------

    def open(self, mode: str = "r", driver=None, comm=None) -> FileHandler:
        """Use this as a context manager in a `with` statement.
        Purpose: keeping the file open to directly access/edit something in the
        HDF5 file of this simulation.

        Args:
            mode (`str`): file mode (see h5py docs)
            driver (`str`): file driver (see h5py docs)
            comm (`str`): mpi communicator
        """
        return self._file(mode, driver, comm)

    @property
    def mesh(self) -> Tuple[np.ndarray, np.ndarray]:
        """Return coordinates and connectivity of default mesh.

        Returns:
            Tuple of np.arrays (coordinates, connectivity)
        """
        return self.get_mesh()

    @with_file_open("r")
    def get_mesh(self, mesh_name: str = None) -> Tuple[np.ndarray, np.ndarray]:
        """Return coordinates and connectivity. Currently returns numpy arrays.

        Args:
            mesh_name (`str`): optional, name of mesh to read (default = mesh)
        Returns:
            Tuple of np.arrays (coordinates, connectivity)
        """
        if mesh_name is None:
            mesh_name = self._default_mesh

        # Raise an error if the mesh is not found
        if (self._mesh_location.split("/")[0] not in self._file.keys()) or (
            mesh_name not in self._file[self._mesh_location].keys()
        ):
            raise KeyError(f"Mesh location {self._mesh_location} not found in file.")

        mesh = self.meshes[mesh_name]
        return mesh.coordinates, mesh.connectivity

    @property
    @deprecated("Use `data.info` instead")
    @with_file_open("r")
    def data_info(self) -> pd.DataFrame:
        """View the data stored.

        Returns:
            :class:`pd.DataFrame`
        """
        tmp_dictionary = dict()
        for data in self.data:
            steps = len(data)
            shape = data.obj["0"].shape
            dtype = data.obj["0"].dtype
            tmp_dictionary[data._name] = {
                "dtype": dtype,
                "shape": shape,
                "steps": steps,
            }
        return pd.DataFrame.from_dict(tmp_dictionary)

    @property
    @with_file_open("r")
    def git(self) -> dict:
        """Get Git information.

        Returns:
            :class:`dict` with different repositories.
        """
        if "git" not in self._file.keys():
            return "Sorrrry, no git information stored :()"
        grp = self._file["git"]
        tmp_dict = {}
        for repo in grp.keys():
            tmp_dict[repo] = grp[repo][()].decode("utf8")
        return tmp_dict

    def get_data_interpolator(self, field: str, step: int):
        """Get Linear interpolator for data field at step. Uses the linked mesh.

        Args:
            name (`str`): name of the data field
            step (`int`): step
        Returns:
            :class:`scipy.interpolate.LinearNDInterpolator`
        """
        from scipy.interpolate import LinearNDInterpolator

        return LinearNDInterpolator(
            self.data[field].mesh.coordinates, self.data[field].at_step(step)
        )

    @with_file_open()
    def show_h5tree(self) -> None:
        """Print the tree inside the h5 file."""
        # print('\U00002B57 ' + os.path.basename(self.h5file))
        print("\U0001F43C " + os.path.basename(self.h5file))
        utilities.h5_tree(self._file.file_object)



# This file is part of bamboost, a Python library built for datamanagement
# using the HDF5 file format.
#
# https://gitlab.ethz.ch/compmechmat/research/libs/dbmanager
#
# Copyright 2023 Flavio Lorez and contributors
#
# There is no warranty for this code
from __future__ import annotations

import logging
import numbers
import os
import pkgutil
import shutil
import uuid
from ctypes import ArgumentError
from typing import Generator, Union

import h5py
import pandas as pd


# forward declaration
class Manager:
    pass


from . import index
from .common.file_handler import open_h5file
from .common.mpi import MPI
from .index import DatabaseTable, IndexAPI, config
from .simulation import Simulation
from .simulation_writer import SimulationWriter

__all__ = ["Manager", "ManagerFromUID", "ManagerFromName"]

log = logging.getLogger(__name__)


class ManagerFromUID(object):
    """Get a database by its UID. This is used for autocompletion in ipython."""

    def __init__(self) -> None:
        # or [] to circumvent Null type (MPI)
        ids = IndexAPI().fetch(f"SELECT id, path FROM dbindex") or []
        self.completion_keys = tuple(
            [f'{key} - {"..."+val[-25:] if len(val)>=25 else val}' for key, val in ids]
        )

    def _ipython_key_completions_(self):
        return self.completion_keys

    def __getitem__(self, key) -> Manager:
        key = key.split()[0]  # take only uid
        return Manager(uid=key, create_if_not_exist=False)


class ManagerFromName(object):
    """Get a database by its path/name. This is used for autocompletion in ipython."""

    def __init__(self) -> None:
        paths = IndexAPI().fetch("SELECT path FROM dbindex") or []
        self.completion_keys = tuple(paths)

    def _ipython_key_completions_(self):
        return self.completion_keys

    def __getitem__(self, key) -> Manager:
        return Manager(key, create_if_not_exist=False)


class Manager:
    """View of database.

    Args:
        path (`str`): path to the directory of the database. If doesn't exist,
            a new database will be created.
        comm (`MPI.Comm`): MPI communicator
        uid: UID of the database

    Attributes:
        FIX_DF: If False, the dataframe of the database is reconstructed every
            time it is accessed.
        fromUID: Access a database by its UID
        fromName: Access a database by its path/name

    Example:
        >>> db = Manager("path/to/db")
        >>> db.df # DataFrame of the database
    """

    FIX_DF = True
    fromUID: ManagerFromUID = ManagerFromUID()
    fromName: ManagerFromName = ManagerFromName()

    def __init__(
        self,
        path: str = None,
        comm: MPI.Comm = MPI.COMM_WORLD,
        uid: str = None,
        create_if_not_exist: bool = True,
    ):
        # provided uid has precedence
        if uid is not None:
            path = self._index.get_path(uid.upper())
            path = comm.bcast(path, root=0)
        self.path = path
        self.comm = comm

        # check if path exists
        if not os.path.isdir(path):
            if not create_if_not_exist:
                raise NotADirectoryError("Specified path is not a valid path.")
            log.info(f"Created new database ({path})")
            self._make_new(path)

        # retrieve the UID of the database from the id file
        # if not found, a new one is generated
        self.UID = uid or self._retrieve_uid()

        # Update the SQL table for the database
        try:
            with self._index.open():
                self._index.insert_path(self.UID, self.path)
                self._table.create_database_table()
                self._table.sync()
        except index.Error as e:
            log.warning(f"index error: {e}")

    def __getitem__(self, key: Union[str, int]) -> Simulation:
        """Returns the simulation in the specified row of the dataframe.

        Args:
            key: The simulation identifier (`str`) or the row index (`int`).
        Returns:
            The selected simulation object.
        """
        if isinstance(key, str):
            return self.sim(key)
        else:
            return self.sim(self.df.loc[key, "id"])

    def _repr_html_(self) -> str:
        """HTML repr for ipython/notebooks. Uses string replacement to fill the
        template code.
        """
        html_string = pkgutil.get_data(__name__, "html/manager.html").decode()
        icon = pkgutil.get_data(__name__, "html/icon.txt").decode()
        return (
            html_string.replace("$ICON", icon)
            .replace("$db_path", self.path)
            .replace("$db_uid", self.UID)
            .replace("$db_size", str(len(self)))
        )

    def __len__(self) -> int:
        return len(self.all_uids)

    def __iter__(self) -> Generator[Simulation]:
        for sim in self.sims():
            yield sim

    def _ipython_key_completions_(self):
        return self.all_uids

    @property
    def _index(self) -> IndexAPI:
        """The index which contains this database."""
        return IndexAPI()

    @property
    def _table(self) -> DatabaseTable:
        """The table in the sql database for this database."""
        return self._index.get_database_table(self.UID)

    def _retrieve_uid(self) -> str:
        """Get the UID of this database from the file tree."""
        try:
            return index.get_uid_from_path(self.path)
        except FileNotFoundError:
            pass

        log.warning("Database exists but no UID found. Generating new UID.")
        return self._make_new(self.path)

    def _make_new(self, path) -> str:
        """Initialize a new database."""
        from datetime import datetime

        # Create directory for database
        os.makedirs(path, exist_ok=True)

        # Assign a unique id to the database
        self.UID = f"{uuid.uuid4().hex[:10]}".upper()
        uid_file = os.path.join(path, f".BAMBOOST-{self.UID}")
        with open(uid_file, "a") as f:
            f.write(self.UID + "\n")
            f.write(f'Date of creation: {datetime.now().strftime("%d/%m/%Y %H:%M:%S")}')
        os.chmod(uid_file, 0o444)  # read only for uid file

        log.info(f"Registered new database (uid = {self.UID})")
        self._index.insert_path(self.UID, path)
        return self.UID

    @property
    def all_uids(self) -> list:
        if not self.FIX_DF or not hasattr(self, "_all_uids"):
            self._all_uids = self._get_uids()
        return self._all_uids

    @all_uids.setter
    def all_uids(self, value: set | list):
        self._all_uids = value

    def _get_uids(self) -> list:
        """Get all simulation names in the database."""
        all_uids = list()
        for dir in os.listdir(self.path):
            if not os.path.isdir(os.path.join(self.path, dir)):
                continue
            if any(
                [i.endswith(".h5") for i in os.listdir(os.path.join(self.path, dir))]
            ):
                all_uids.append(dir)
        return all_uids

    @property
    def df(self) -> pd.DataFrame:
        """View of the database and its parametric space.

        Returns:
            :class:`pd.DataFrame`
        """
        if not hasattr(self, "_dataframe"):
            return self.get_view()
        if self.FIX_DF and self._dataframe is not None:
            return self._dataframe
        return self.get_view()

    def _get_parameters_for_uid(
        self, uid: str, include_linked_sims: bool = False
    ) -> dict:
        """Get the parameters for a given uid.

        Args:
            uid (`str`): uid of the simulation
            include_linked_sims (`bool`): if True, include the parameters of linked sims
        """
        h5file_for_uid = os.path.join(self.path, uid, f"{uid}.h5")
        tmp_dict = dict()

        with open_h5file(h5file_for_uid, "r") as f:
            if "parameters" in f.keys():
                tmp_dict.update(f["parameters"].attrs)
            if "additionals" in f.keys():
                tmp_dict.update({"additionals": dict(f["additionals"].attrs)})
            tmp_dict.update(f.attrs)

        if include_linked_sims:
            for linked, full_uid in self.sim(uid).links.attrs.items():
                sim = Simulation.fromUID(full_uid)
                tmp_dict.update(
                    {f"{linked}.{key}": val for key, val in sim.parameters.items()}
                )
        return tmp_dict

    def get_view(self, include_linked_sims: bool = False) -> pd.DataFrame:
        """View of the database and its parametric space. Read from the sql
        database. If `include_linked_sims` is True, the individual h5 files are
        scanned.

        Args:
            include_linked_sims: if True, include the parameters of linked sims

        Examples:
            >>> db.get_view()
            >>> db.get_view(include_linked_sims=True)
        """
        if include_linked_sims:
            return self.get_view_from_hdf_files(include_linked_sims=include_linked_sims)

        try:
            with self._table.open():
                self._table.sync()
                df = self._table.read_table()
        except index.Error as e:
            log.warning(f"index error: {e}")
            return self.get_view_from_hdf_files(include_linked_sims=include_linked_sims)

        if df.empty:
            return df
        df["time_stamp"] = pd.to_datetime(df["time_stamp"])

        # Sort dataframe columns
        columns_start = ["id", "notes", "status", "time_stamp"]
        self._dataframe = df[[*columns_start, *df.columns.difference(columns_start)]]

        opts = config.get("options", {})
        if "sort_table_key" in opts:
            self._dataframe.sort_values(
                opts.get("sort_table_key", "id"),
                ascending=opts.get("sort_table_order", "asc") == "asc",
                inplace=True,
            )
        return self._dataframe

    def get_view_from_hdf_files(
        self, include_linked_sims: bool = False
    ) -> pd.DataFrame:
        """View of the database and its parametric space. Read from the h5
        files metadata.

        Args:
            include_linked_sims: if True, include the parameters of linked sims
        """
        all_uids = self._get_uids()
        data = list()

        for uid in all_uids:
            tmp_dict = self._get_parameters_for_uid(
                uid, include_linked_sims=include_linked_sims
            )
            data.append(tmp_dict)

        df = pd.DataFrame.from_records(data)
        if df.empty:
            return df
        df["time_stamp"] = pd.to_datetime(df["time_stamp"])

        # Sort dataframe columns
        columns_start = ["id", "notes", "status", "time_stamp"]
        self._dataframe = df[[*columns_start, *df.columns.difference(columns_start)]]
        return self._dataframe

    @property
    def data_info(self) -> pd.DataFrame:
        """Return view of stored data for all simulations

        Returns:
            :class:`pd.DataFrame`
        """
        data = list()
        for uid in self.all_uids:
            h5file_for_uid = os.path.join(self.path, uid, f"{uid}.h5")
            with open_h5file(h5file_for_uid, "r") as file:
                try:
                    tmp_dict = dict()
                    tmp_dict = {
                        key: (
                            len(file[f"data/{key}"]),
                            file[f"data/{key}/0"].shape,
                            file[f"data/{key}/0"].dtype,
                        )
                        for key in file["data"].keys()
                    }
                except KeyError:
                    tmp_dict = dict()
                data.append(tmp_dict)
        return pd.DataFrame.from_records(data)

    def sim(
        self,
        uid,
        return_writer: bool = False,
        writer_type: SimulationWriter = SimulationWriter,
    ) -> Simulation:
        """Get an existing simulation with uid. Same as accessing with `db[uid]` directly.

        Args:
            uid (`str`): unique identifier
            return_writer: if true, return `SimulationWriter`, otherwise
                return `Simulation`
            writer_type: Optionally, you can specify a custom writer type to return.

        Returns:
            :class:`~bamboost.simulation.Simulation`
        """
        if return_writer:
            return writer_type(uid, self.path, self.comm)
        return Simulation(uid, self.path, self.comm, _db_id=self.UID)

    def sims(
        self,
        select: pd.Series = None,
        sort: str = None,
        reverse: bool = False,
        exclude: set = None,
        return_writer: bool = False,
    ) -> list:
        """Get all simulations in a list. Optionally, get all simulations matching the
        given selection using pandas.

        Args:
            select (`pd.Series`): pandas boolean series
            sort (`str`): Optionally sort the list with this keyword
            reverse (`bool`): swap sort direction
            exclude (`list[str]`): sims to exclude
            return_writer: if true, return `SimulationWriter`, otherwise
                return `Simulation`

        Returns:
            A list of `:class:~bamboost.simulation.Simulation` objects

        Examples:
            >>> db.sims(select=db.df["status"] == "finished", sort="time_stamp")
        """
        if select is not None:
            id_list = self.df[select]["id"].values
        else:
            id_list = self.all_uids
        if exclude is not None:
            exclude = list([exclude]) if isinstance(exclude, str) else exclude
            id_list = [id for id in id_list if id not in exclude]

        existing_sims = [self.sim(uid, return_writer) for uid in id_list]

        if sort is None:
            return existing_sims
        else:
            return sorted(
                existing_sims, key=lambda s: s.parameters[sort], reverse=reverse
            )

    def create_simulation(
        self,
        uid: str = None,
        parameters: dict = None,
        skip_duplicate_check: bool = False,
        prefix: str = None,
    ) -> SimulationWriter:
        """Get a writer object for a new simulation. This is written for paralell use
        as it is likely that this may be used in an executable, creating multiple runs
        for a parametric space, which may be run in paralell.

        Args:
            uid (`str`): The name/uid for the simulation. If not specified, a random id
                will be assigned.
            parameters (`dict`): Parameter dictionary. If provided, the parameters will be
                checked against the existing sims for duplication. Otherwise, they may be
                specified later with :func:`~bamboost.simulation.SimulationWriter.add_parameters`.
            skip_duplicate_check (`bool`): if True, the duplicate check is skipped.
            prefix (`str`): Prefix for the uid. If not specified, no prefix is used.

        Returns:
            sim (:class:`~bamboost.simulation.SimulationWriter`)

        Examples:
            >>> db.create_simulation(parameters={"a": 1, "b": 2})
            apple

            >>> db.create_simulation(uid="my_sim", parameters={"a": 1, "b": 2}, prefix="test")
        """
        if parameters and not skip_duplicate_check:
            go_on, uid = self._check_duplicate(parameters, uid)
            if not go_on:
                print("Aborting by user desire...")
                return None

        if self.comm.rank == 0:
            if not uid:
                uid = uuid.uuid4().hex[:8]  # Assign random unique identifier
            if isinstance(prefix, str) and prefix != "":
                uid = "_".join([prefix, uid])
        uid = self.comm.bcast(uid, root=0)

        # Create directory and h5 file
        if self.comm.rank == 0:
            os.makedirs(os.path.join(self.path, uid), exist_ok=True)
            path_to_h5_file = os.path.join(self.path, uid, f"{uid}.h5")
            if os.path.exists(path_to_h5_file):
                os.remove(path_to_h5_file)
            h5py.File(path_to_h5_file, "a").close()  # create file

        new_sim = SimulationWriter(uid, self.path, self.comm)
        new_sim.initialize()  # sets metadata and status
        # add the id to the (fixed) _all_uids list
        if hasattr(self, "_all_uids"):
            self._all_uids.append(new_sim.uid)
        if parameters:
            new_sim.add_parameters(parameters)
        return new_sim

    def remove(self, uid: str) -> None:
        """CAUTION, DELETING DATA. Remove the data of a simulation.

        Args:
            uid (`str`): uid
        """
        shutil.rmtree(os.path.join(self.path, uid))

    def _check_duplicate(self, parameters: dict, uid: str) -> tuple:
        """Checking whether the parameters dictionary exists already.
        May need to be improved...

        Args:
            parameters (`dict`): parameter dictionary to check for
            uid (`str`): uid
        Returns:
            Tuple(Bool, uid) wheter to continue and with what uid.
        """
        duplicates = list()

        for _uid in self.all_uids:
            with open_h5file(
                os.path.join(os.path.join(self.path, _uid), f"{_uid}.h5"), "r"
            ) as f:
                if "parameters" not in f.keys():
                    continue

                tmp_dict = dict()
                tmp_dict.update(f["parameters"].attrs)

                if tmp_dict == parameters:
                    duplicates.append((_uid, "equal"))
                    continue

                shared_keys = tmp_dict.keys() & parameters.keys()
                if {key: tmp_dict[key] for key in shared_keys} == {
                    key: parameters[key] for key in shared_keys
                }:
                    duplicates.append((_uid, "shared_equal"))
                    continue

        if not duplicates:
            return True, uid

        # What should be done?
        print(
            f"The parameter space may already exist. Here are the duplicates:",
            flush=True,
        )
        print(self.df[self.df["id"].isin([i[0] for i in duplicates])], flush=True)

        prompt = input(
            "Replace first duplicate ('r'), Create with altered uid (`c`), "
            + "Create new with new id (`n`), Abort (`a`): "
        )
        if prompt == "r":
            self.remove(duplicates[0][0])
            return True, uid
        if prompt == "a":
            return False, uid
        if prompt == "n":
            return True, uid
        if prompt == "c":
            return True, self._generate_subuid(duplicates[0][0].split(".")[0])

        raise ArgumentError("Answer not valid! Aborting")

    def _generate_subuid(self, uid_base: str) -> str:
        """Return a new sub uid for the base uid.
        Following the following format: `base_uid.1`

        Args:
            uid_base (`str`): base uid for which to find the next subid.
        Returns:
            New uid string
        """
        uid_list = [uid for uid in self.all_uids if uid.startswith(uid_base)]
        subiterator = max(
            [int(id.split(".")[1]) for id in uid_list if len(id.split(".")) > 1] + [0]
        )
        return f"{uid_base}.{subiterator+1}"

    def global_fields_in_all(self) -> list:
        """Get a list of all global fields in all simulations.

        Returns:
            List of global fields
        """
        fields = set()
        for sim in self:
            try:
                fields.update(sim.globals.columns)
            except KeyError:
                continue

        return fields

    def get_parameters(self) -> dict:
        """Get the parameters used in this database.

        Returns:
            Dictionary of parameters with it's count, range, and type. Sorted by count.
        """
        parameters = dict()
        for sim in self:
            for key, val in sim.parameters.items():
                if key not in parameters:
                    range = (val, val) if isinstance(val, numbers.Number) else None
                    parameters[key] = {"range": range, "count": 1, "type": type(val)}
                else:
                    if isinstance(val, numbers.Number):
                        parameters[key]["range"] = (
                            min(parameters[key]["range"][0], val),
                            max(parameters[key]["range"][1], val),
                        )
                    parameters[key]["count"] += 1
                    parameters[key]["type"] = type(val)
        return dict(
            sorted(parameters.items(), key=lambda x: x[1]["count"], reverse=True)
        )



from test_manager import temp_manager
import numpy as np
import pytest


@pytest.mark.parametrize("dtype,array", [['<f4', np.array([1, 2], dtype=int)],
                                         ["int32", np.array([1, 2])],
                                         ])
def test_field_dtype_forced(temp_manager, dtype, array):
    sim = temp_manager.create_simulation()
    uid = sim.uid
    sim.add_field("test", array, dtype=dtype)

    sim.finish_step()
    sim.add_field("test", 2 * array, dtype=dtype)

    sim.finish_sim()

    sim = temp_manager.sim(uid)

    assert sim.data["test"].dtype == dtype


@pytest.mark.parametrize("dtype,value", [['<f4', 1],
                                         ["int32", 1],
                                         ])
def test_global_field_dtype_forced(temp_manager, dtype, value):
    sim = temp_manager.create_simulation()
    uid = sim.uid
    sim.add_global_field("test", value, dtype=dtype)

    sim.finish_step()
    sim.add_global_field("test", 2 * value, dtype=dtype)

    sim.finish_sim()

    sim = temp_manager.sim(uid)

    assert sim.globals["test"].dtype == dtype


@pytest.mark.parametrize("array", [np.array([1., 2.]),
                                   np.array([1, 2], dtype="int"),
                                   ])
def test_field_dtype_inferred(temp_manager, array):
    sim = temp_manager.create_simulation()
    uid = sim.uid
    sim.add_field("test", array)

    sim.finish_step()
    sim.add_field("test", 2 * array)

    sim.finish_sim()

    sim = temp_manager.sim(uid)

    # The dtypes are equivalent (although they may have another string representation)
    assert np.dtype(sim.data["test"].dtype) == array.dtype


@pytest.mark.parametrize("val", [1., 1, ])
def test_global_field_dtype_inferred(temp_manager, val):
    sim = temp_manager.create_simulation()
    uid = sim.uid
    sim.add_global_field("test", val)

    sim.finish_step()
    sim.add_global_field("test", 2 * val)

    sim.finish_sim()

    sim = temp_manager.sim(uid)

    # The dtypes are equivalent (although they may have another string representation)
    assert np.dtype(sim.globals["test"].dtype) == np.array([val]).dtype


@pytest.mark.parametrize("array", [np.array([1., 2.]),
                                   np.array([1, 2], dtype="int"),
                                   ])
def test_userdata_dtype_inferred(temp_manager, array):
    sim = temp_manager.create_simulation()
    uid = sim.uid
    sim.userdata.add_dataset("test", array)

    sim.finish_step()
    sim.userdata.add_dataset("test", 2 * array)

    sim.finish_sim()

    sim = temp_manager.sim(uid)

    # The dtypes are equivalent (although they may have another string representation)
    assert np.dtype(sim.userdata["test"].dtype) == array.dtype


@pytest.mark.parametrize("dtype,array", [['<f4', np.array([1, 2], dtype=int)],
                                         ["int32", np.array([1, 2])],
                                         ])
def test_userdata_dtype_forced(temp_manager, dtype, array):
    sim = temp_manager.create_simulation()
    uid = sim.uid
    sim.userdata.add_dataset("test", array, dtype=dtype)

    sim.finish_step()
    sim.userdata.add_dataset("test", 2 * array, dtype=dtype)

    sim.finish_sim()

    sim = temp_manager.sim(uid)

    assert sim.userdata["test"].dtype == dtype
